"""Return elements chosen from `x` or `y` depending on `condition`."""
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.where)
def where(condition, *args):
    """
    Return elements chosen from `x` or `y` depending on `condition`.

    .. note::
        When only `condition` is provided, this function is a shorthand for
        ``np.asarray(condition).nonzero()``. Using `nonzero` directly should be
        preferred, as it behaves correctly for subclasses. The rest of this
        documentation covers only the case where all three arguments a re
        provided.

    Args:
        condition (numpy.ndarray, bool):
            Where True, yield `x`, otherwise yield `y`.
        x (numpoly.ndpoly): array_like
            Values from which to choose. `x`, `y` and `condition` need to be
            broadcastable to some shape.

    Returns:
        (numpoly.ndpoly):
            An array with elements from `x` where `condition` is True, and
            elements from `y` elsewhere.

    Examples:
        >>> poly = numpoly.symbols("x")*numpy.arange(4)
        >>> poly
        polynomial([0, x, 2*x, 3*x])
        >>> numpoly.where([1, 0, 1, 0], 7, 2*poly)
        polynomial([7, 2*x, 7, 6*x])
        >>> numpoly.where(poly, 2*poly, 4)
        polynomial([4, 2*x, 4*x, 6*x])
        >>> numpoly.where(poly)
        (array([1, 2, 3]),)

    """
    if isinstance(condition, numpoly.ndpoly):
        condition = numpy.any(condition.coefficients, 0).astype(bool)
    if not args:
        return numpy.where(condition)

    poly1, poly2 = numpoly.align_polynomials(*args)
    out = numpy.where(condition, poly1.values, poly2.values)
    out = numpoly.polynomial(out, names=poly1.indeterminants)
    return out



"""Apply a function repeatedly over multiple axes."""
from functools import wraps
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.apply_over_axes)
def apply_over_axes(func, a, axes):
    """
    Apply a function repeatedly over multiple axes.

    `func` is called as `res = func(a, axis)`, where `axis` is the first
    element of `axes`.  The result `res` of the function call must have either
    the same dimensions as `a` or one less dimension.  If `res` has one less
    dimension than `a`, a dimension is inserted before `axis`.  The call to
    `func` is then repeated for each axis in `axes`, with `res` as the first
    argument.

    Args:
        func (Callable):
            This function must take two arguments, `func(a, axis)`.
        a (numpoly.ndpoly):
            Input array.
        axes (numpy.ndarray, int):
            Axes over which `func` is applied; the elements must be integers.

    Returns:
        (numpoly.ndpoly):
            The output array.  The number of dimensions is the same as `a`, but
            the shape can be different.  This depends on whether `func` changes
            the shape of its output with respect to its input.

    Examples:
        >>> polynomial = numpoly.symbols("x y z")*numpy.arange(24).reshape(2, 4, 3)
        >>> numpoly.apply_over_axes(numpoly.sum, polynomial, 1)
        polynomial([[[18*x, 22*y, 26*z]],
        <BLANKLINE>
                    [[66*x, 70*y, 74*z]]])
        >>> numpoly.apply_over_axes(numpoly.sum, polynomial, [0, 2])
        polynomial([[[16*z+14*y+12*x],
                     [22*z+20*y+18*x],
                     [28*z+26*y+24*x],
                     [34*z+32*y+30*x]]])

    """
    @wraps(func)
    def wrapper_func(array, axis):
        """Wrap func function."""
        # Align indeterminants in case slicing changed them
        array = numpoly.polynomial(array, names=a.indeterminants)
        array, _ = numpoly.align.align_indeterminants(array, a.indeterminants)
        # Evaluate function
        out = func(array, axis=axis)
        # Restore indeterminants in case func changed them.
        out, _ = numpoly.align.align_indeterminants(out, a.indeterminants)
        return out

    # Initiate wrapper
    a = numpoly.aspolynomial(a)
    out = numpy.apply_over_axes(wrapper_func, a=a.values, axes=axes)
    return out



"""Apply a function to 1-D slices along the given axis."""
from functools import wraps
import numpy
import numpoly

from ..dispatch import implements


@implements(numpy.apply_along_axis)
def apply_along_axis(func1d, axis, arr, *args, **kwargs):
    """
    Apply a function to 1-D slices along the given axis.

    Execute `func1d(a, *args)` where `func1d` operates on 1-D arrays and `a` is
    a 1-D slice of `arr` along `axis`.

    This is equivalent to (but faster than) the following use of `ndindex` and
    `s_`, which sets each of ``ii``, ``jj``, and ``kk`` to a tuple of indices::

        Ni, Nk = a.shape[:axis], a.shape[axis+1:]
        for ii in ndindex(Ni):
            for kk in ndindex(Nk):
                f = func1d(arr[ii + s_[:,] + kk])
                Nj = f.shape
                for jj in ndindex(Nj):
                    out[ii + jj + kk] = f[jj]

    Equivalently, eliminating the inner loop, this can be expressed as::

        Ni, Nk = a.shape[:axis], a.shape[axis+1:]
        for ii in ndindex(Ni):
            for kk in ndindex(Nk):
                out[ii + s_[...,] + kk] = func1d(arr[ii + s_[:,] + kk])

    Args:
        func1d (Callable[[numpoly.ndpoly], Any]):
            This function should accept 1-D arrays. It is applied to 1-D slices
            of `arr` along the specified axis.
        axis (int):
            Axis along which `arr` is sliced.
        arr (numpoly.ndpoly):
            Input array.
        args:
            Additional arguments to `func1d`.
        kwargs:
            Additional named arguments to `func1d`.

    Returns:
        out (numpoly.ndpoly):
            The output array. The shape of `out` is identical to the shape of
            `arr`, except along the `axis` dimension. This axis is removed, and
            replaced with new dimensions equal to the shape of the return value
            of `func1d`. So if `func1d` returns a scalar `out` will have one
            fewer dimensions than `arr`.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> b = numpoly.polynomial([[1, 2, 3*x], [3, 6*y, 6], [2, 7, 9]])
        >>> numpoly.apply_along_axis(numpoly.mean, 0, b)
        polynomial([2.0, 3.0+2.0*y, 5.0+x])
        >>> numpoly.apply_along_axis(numpoly.mean, 1, b)
        polynomial([1.0+x, 3.0+2.0*y, 6.0])

    """
    collection = list()

    @wraps(func1d)
    def wrapper_func(array):
        """Wrap func1d function."""
        # Align indeterminants in case slicing changed them
        array = numpoly.polynomial(array, names=arr.indeterminants)
        array, _ = numpoly.align.align_indeterminants(array, arr.indeterminants)

        # Evaluate function
        out = func1d(array, *args, **kwargs)

        # Restore indeterminants in case func1d changed them.
        out, _ = numpoly.align.align_indeterminants(out, arr.indeterminants)

        # Return dummy index integer value that will be replaced with
        # polynomials afterwards.
        ret_val = len(collection)*numpy.ones(out.shape, dtype=int)
        collection.append(out)
        return ret_val

    # Initiate wrapper
    arr = numpoly.aspolynomial(arr)
    out = numpy.apply_along_axis(wrapper_func, axis=axis, arr=arr.values)

    # align indeterminants and exponents
    polynomials = numpoly.align.align_indeterminants(*collection)
    polynomials = numpoly.align.align_exponents(*polynomials)
    dtype = numpoly.common_type(*polynomials)

    # Store results into new array
    ret_val = numpoly.ndpoly(
        exponents=polynomials[0].exponents,
        shape=out.shape,
        names=polynomials[0].indeterminants,
        dtype=dtype,
    ).values
    for idx, polynomial in enumerate(polynomials):
        ret_val[out == idx] = polynomial.values

    return numpoly.aspolynomial(
        ret_val, dtype=dtype, names=polynomials[0].indeterminants)



"""Polynomial functionality."""
from .call import call
from .decompose import decompose
from .derivative import diff, gradient, hessian
from .isconstant import isconstant
from .monomial import monomial, bindex
from .divide import poly_divide, poly_divmod, poly_remainder
from .symbols import symbols
from .tonumpy import tonumpy

__all__ = ("bindex", "call", "decompose", "diff", "gradient", "hessian",
           "isconstant", "monomial", "poly_divide", "poly_divmod",
           "poly_remainder", "symbols", "tonumpy")



"""Return element-wise quotient and remainder simultaneously."""
from __future__ import division
import numpy
import numpoly

from ...dispatch import implements_function


@implements_function(numpy.divmod)
def poly_divmod(dividend, divisor, out=(None, None), where=True, **kwargs):
    """
    Return element-wise quotient and remainder simultaneously.

    ``numpoly.divmod(x, y)`` is equivalent to ``(x / y, x % y)``, but faster
    because it avoids redundant work. It is used to implement the Python
    built-in function ``divmod`` on Numpoly arrays.

    Notes:
        Unlike numbers, this returns the polynomial division and polynomial
        remainder. This means that this function is _not_ backwards compatible
        with ``numpy.divmod`` for constants. For example:
        ``numpy.divmod(11, 2) == (5, 1)`` while
        ``numpoly.divmod(11, 2) == (5.5, 0)``.

    Args:
        dividend (numpoly.ndpoly):
            The array being divided.
        divisor (numpoly.ndpoly):
            Array that that will divide the dividend.
        out (Tuple[Optional[numpoly.ndpoly], Optional[numpoly.ndpoly]]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (bool, numpy.ndarray):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (Tuple[numpoly.ndpoly, numpoly.ndpoly]):
            Element-wise quotient and remainder resulting from
            floor division. This is a scalar if both `x1` and `x2`
            are scalars.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> denominator = [x*y**2+2*x**3*y**2, -2+x*y**2]
        >>> numerator = -2+x*y**2
        >>> floor, remainder = numpoly.poly_divmod(denominator, numerator)
        >>> floor
        polynomial([1.0+2.0*x**2, 1.0])
        >>> remainder
        polynomial([2.0+4.0*x**2, 0.0])
        >>> floor*numerator+remainder
        polynomial([x*y**2+2.0*x**3*y**2, -2.0+x*y**2])

    """
    assert where is True, "changing 'where' is not supported."
    dividend, divisor = numpoly.align_polynomials(dividend, divisor)

    if not dividend.shape:
        floor, remainder = poly_divmod(
            dividend.flatten(), divisor.flatten(), out=out, where=where, **kwargs)
        return floor[0], remainder[0]

    quotient = numpoly.zeros(dividend.shape)
    while True:

        candidates = get_division_candidate(dividend, divisor)
        if candidates is None:
            break
        idx1, idx2, include = candidates

        # construct candidate
        candidate = dividend.coefficients[idx1]/numpy.where(
            include, divisor.coefficients[idx2], 1)
        exponent_diff = dividend.exponents[idx1]-divisor.exponents[idx2]
        candidate = candidate*numpoly.prod(divisor.indeterminants**exponent_diff, 0)

        # iterate division algorithm
        quotient = numpoly.add(
            quotient, numpoly.where(include, candidate, 0), **kwargs)
        dividend = numpoly.subtract(
            dividend, numpoly.where(include, divisor*candidate, 0), **kwargs)
        dividend, divisor = numpoly.align_polynomials(dividend, divisor)

    return quotient, dividend


def get_division_candidate(x1, x2):
    """
    Find the next exponent candidate pair in the iterative subtraction process.

    Args:
        x1 (numpoly.ndpoly):
            The array being divided.
        x2 (numpoly.ndpoly):
            Array that that will divide the dividend.

    Returns:
        (Optional[Tuple[int, int, numpy.ndarray]):
            Indices to the exponent candidate in respectively `x1` and `x2`
            that should be used in the divide process, and an boolean array to
            indicate for which coefficients these candidates are valid.

    """
    # Look for exponent candidates among divisors
    for idx2 in reversed(numpy.lexsort(x2.exponents.T)):
        exponent2 = x2.exponents[idx2]

        # Include coefficients where idx2 is non-zero and any potential
        # candidates that is a better fit has coefficient zero. Exponent needs
        # to be the biggest one around.
        include2 = numpy.ones(x2.shape, dtype=bool)
        for idx, exponent in enumerate(x2.exponents):
            if numpy.all(exponent2 <= exponent):
                include2 &= (x2.coefficients[idx] == 0) ^ (idx == idx2)
        if not numpy.any(include2):
            continue

        # Look for exponent candidates among dividend
        for idx1 in reversed(numpy.lexsort(x1.exponents.T)):
            exponent1 = x1.exponents[idx1]

            # Skip situations where numerator has larger exponent than the
            # denominator.
            if numpy.any(exponent1 < exponent2):
                continue

            # Skip if all relevant coefficients of interest are zero.
            include1 = x1.coefficients[idx1] != 0
            include = include1 & include2
            if not numpy.any(include):
                continue

            return idx1, idx2, include

    # No valid candidate pair found; Assume we are done.
    return None



"""Return element-wise remainder of polynomial division."""
import numpy
import numpoly

from ...dispatch import implements_function
from .divmod import poly_divmod


@implements_function(numpy.remainder)
def poly_remainder(x1, x2, out=None, where=True, **kwargs):
    """
    Return element-wise remainder of polynomial division.

    Args:
        x1 (numpoly.ndpoly):
            Dividend array.
        x2 (numpoly.ndpoly):
            Divisor array. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]):
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Union[bool, numpy.ndarray]):
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        (numpoly.ndpoly):
            The element-wise remainder of the quotient
            ``floor_divide(x1, x2)``. This is a scalar if both `x1` and `x2`
            are scalars.

    Notes:
        Unlike numbers, this returns the polynomial division and polynomial
        remainder. This means that this function is _not_ backwards compatible
        with ``numpy.remainder`` for constants. For example:
        ``numpoly.remainder(11, 2) == 1`` while
        ``numpoly.poly_remainder(11, 2) == 0``.

    Examples:
        >>> x, y = numpoly.symbols("x y")
        >>> denominator = [x*y**2+2*x**3*y**2, -2+x*y**2]
        >>> numerator = -2+x*y**2
        >>> numpoly.poly_remainder(denominator, numerator)
        polynomial([2.0+4.0*x**2, 0.0])

    """
    dividend, remainder = poly_divmod(x1, x2, out=out, where=where, **kwargs)
    return remainder
